use crate::patterns::{Pattern, PixelColor};
use crate::{impl_pattern_last_iteration, Strip};
use rand::Rng;
use serde::{Deserialize, Serialize};

/// # StarsRandom
///
/// fill strip with background color then color random pixel with other color fading to background color
#[derive(Serialize, Deserialize, Debug, Eq, Hash, PartialEq, Copy, Clone)]
pub struct StarsRandom<const N: usize> {
    pub(crate) color: PixelColor,
    pub(crate) ratio: usize,
    pub(crate) max_iteration: Option<usize>,
    pub(crate) background_color: Option<PixelColor>,
    pub(crate) steps: usize,
    #[serde(default)]
    pub(crate) current_iteration: usize,
    #[serde(skip)]
    strip: Strip<N>,
    #[serde(skip)]
    step: (i8, i8, i8),
}

impl<const N: usize> Pattern for StarsRandom<N> {
    type Strip = Strip<N>;
    fn init(&mut self) -> Option<Self::Strip> {
        if let Some(color) = self.background_color {
            self.strip = Strip::<N>::new(color);
            let red = (self.color.red as i16 - color.red as i16) as i8;
            let green = (color.green as i16 - self.color.green as i16) as i8;
            let blue = (self.color.blue as i16 - color.blue as i16) as i8;
            println!("{} {} {}", red, green, blue);
            self.step = (
                red / self.steps as i8,
                green / self.steps as i8,
                blue / self.steps as i8,
            );
            Some(self.strip)
        } else {
            self.step = (
                -((self.color.red / self.steps as u8) as i8),
                -((self.color.green / self.steps as u8) as i8),
                -((self.color.blue / self.steps as u8) as i8),
            );
            None
        }
    }
    impl_pattern_last_iteration!(StarsRandom);
}

impl<const N: usize> Iterator for StarsRandom<N> {
    type Item = Strip<N>;

    fn next(&mut self) -> Option<Self::Item> {
        if self.is_last_iteration() {
            return None;
        }

        let mut rng = rand::thread_rng();
        for i in 0..N {
            let current_color = self.strip[i];
            if current_color == self.background_color {
                if rng.gen_bool(1.0 / self.ratio as f64) {
                    self.strip[i] = Some(self.color);
                }
            } else {
                let color = current_color.unwrap();
                let background = self.background_color.unwrap_or(PixelColor::default());
                let distance = PixelColor {
                    red: self.step.0.abs() as u8,
                    green: self.step.1.abs() as u8,
                    blue: self.step.2.abs() as u8,
                };
                if color.is_closer(background, distance) {
                    self.strip[i] = self.background_color;
                } else {
                    self.strip[i] = Some(color.delta(self.step));
                }
            }
        }

        self.current_iteration += 1;
        Some(self.strip)
    }
}
